import os
import sys
import time
import zipfile

import click
import json
import shutil
import requests
import tempfile
from colorama import Fore

from . import LOG, USER_DIRECTORIES, __PLUGIN_DATA__
from .adb import Adb
from .apktool import Apktool
from .java import Java
from .arch import ARCH
from .download import ProgressCompleted, ProgressDownloading, ProgressFailed, ProgressUnknown
from .frida import Frida
from .gadget import Gadget
from .injector import Injector
from .plugins import _PluginLoader, main
from .uber_apk_signer import UberApkSigner
from .bundle import Bundle, ConfigArch, ConfigDpi, ConfigLocale
from .download import download_file
from .manifest import Manifest

def _clear_cache(func):
    def inner(*args, **kwargs):
        func(*args, **kwargs)
        shutil.rmtree(USER_DIRECTORIES.user_cache_dir)
        os.makedirs(USER_DIRECTORIES.user_cache_dir)
    return inner


def _install_callback(progress):
    if isinstance(progress, ProgressDownloading):
        sys.stdout.write(
            Fore.GREEN + f'\r[+] Downloading {progress.filename} - {progress.progress:03d}%')
        sys.stdout.flush()
    if isinstance(progress, ProgressFailed):
        sys.stdout.write(
            Fore.RED + f'[*] Download failed with status: {progress.status_code}')
    if isinstance(progress, ProgressCompleted):
        sys.stdout.write('\n')
    if isinstance(progress, ProgressUnknown):
        LOG.info('Downloading {} this might take a while...', progress.filename)


@_clear_cache
def _inject(apk, libraries, include, activity, output, override, use_aapt):
    if not os.path.isfile(apk):
        LOG.error('{} - not such file.', apk)
    for lib in libraries:
        if not os.path.isfile(lib):
            LOG.error('{} - not such file.', lib)
    
    if include:
        for i in include:
            if not os.path.isfile(i):
                LOG.error('Include {} is not a valid file.', include)

    apk_name, ext =  os.path.splitext(apk)
    workdir = os.path.join(USER_DIRECTORIES.user_cache_dir, apk_name)

    libraries = list(libraries)
    entry_lib = libraries[0]

    _, _ext = os.path.splitext(output)
    if not _ext:
        # is file
        tmp_output = output
        if os.path.isdir(output):
            tmp_output = os.path.join(output, f'{apk_name}_patched{ext}' if not override else apk)
        else:
            if tmp_output.endswith(os.sep):
                tmp_output = tmp_output.rsplit(os.sep, 1)[0]
        if os.path.exists(tmp_output):
            timestamp = int(time.time())
            tmp_output = os.path.join(output, f'{apk_name}_patched_{timestamp}{ext}')
        output = tmp_output
    else:
        if os.path.exists(output) and not override:
            LOG.error('{} already exists.', output)

    is_bundle = Bundle.is_bundle(apk)

    output_bundle = None
    bundle_info = None

    if is_bundle:
        LOG.info('Extracting split apks...')
        output_bundle = f'{workdir}_bundle'
        if os.path.exists(output_bundle):
            shutil.rmtree(output_bundle)
        bundle_info = Bundle.extract(apk, output_bundle)

    targets = []

    if bundle_info and bundle_info.configs:
        for config in bundle_info.configs:
            if isinstance(config, ConfigArch):
                for lib in libraries.copy():
                    lib_arch = Injector.guess_arch_from_lib(lib)
                    if lib_arch == config.arch:
                        target_split = os.path.join(output_bundle, config.file_name)
                        targets.append([target_split, lib])
                        libraries.remove(lib)
    for lib in libraries:
        if is_bundle:
            target_split = os.path.join(output_bundle, bundle_info.base_apk)
            targets.append([target_split, lib])
        else:
            targets.append([output, lib])

    Apktool.install(progress_callback=_install_callback)
    LOG.info('Using apktool: {}', Apktool.version())

    base = os.path.join(output_bundle, bundle_info.base_apk) if is_bundle else apk
    # extract base
    LOG.info('Extracting {}', base)
    Apktool.decode(base, force=True, output=workdir, framework_path=f'{workdir}_framework')

    manifest_path = os.path.join(workdir, 'AndroidManifest.xml')
    target_activity = None
    if activity:
        for activities in Manifest.get_activities(manifest_path):
            for i in activities:
                if activity in i:
                    target_activity = i
                    break
    else:
        target_activity = Manifest.get_main_activity(manifest_path)
    if not target_activity:
        LOG.error('{} did not match any activities', activity)
    
    LOG.info('Found target activity {}', target_activity)

    entrypoint = None
    for file in os.listdir(workdir):
        if file.startswith('smali'):
            entrypoint_tmp = os.path.join(
            workdir, file, target_activity.replace('.', '/') + '.smali')
            if os.path.isfile(entrypoint_tmp):
                entrypoint = entrypoint_tmp
                break  
    if not entrypoint:
        LOG.error('Failed to find smali file for activity {}', target_activity)
        
    LOG.info('Injecting smali loader into {}.', target_activity)
    injected = Injector.inject_library(entry_lib, workdir, entrypoint, include, skip_copy=True) # Copying will be done for each target bellow
    if not injected:
        LOG.error('Something wen\'t wrong when injecting into target..')
    LOG.info('Repacking {} -> {}', base if is_bundle else apk, base if is_bundle else output)
    Apktool.build(workdir, output=base if is_bundle else output, framework_path=f'{workdir}_framework', use_aapt2=use_aapt)

    for target in targets:
        apk_target = target[0]
        lib_target = target[1]
        if os.path.exists(workdir):
            shutil.rmtree(workdir)
        os.makedirs(workdir)
        LOG.info('Extracting {}', apk_target)
        with zipfile.ZipFile(apk_target, 'r') as ref:
            ref.extractall(workdir)
        LOG.info('Injecting {} in {}', lib_target, apk_target)
        injected = Injector.inject_library(lib_target, workdir, smali_path=None, extra_files=include)
        if not injected:
            LOG.error('Something wen\'t wrong when injecting into target..')
        LOG.info('Repacking {} -> {}', apk_target, apk_target if is_bundle else output)
        with zipfile.ZipFile(apk_target if is_bundle else output, 'w') as ref:
            for root, dirs, files in os.walk(workdir):
                for file in files:
                    ref.write(os.path.join(root, file), 
                        os.path.relpath(os.path.join(root, file), 
                                        workdir))
    
    apks = []
    if is_bundle:
        apks = [os.path.join(output_bundle, apk) for apk in os.listdir(output_bundle) if apk.endswith('.apk')]
    else:
        apks = [output,]
    
    LOG.info('Signing {}', ' '.join([os.path.basename(apk) for apk in apks]))
    UberApkSigner.install(progress_callback=_install_callback)

    LOG.info('Using uber-apk-signer {}', UberApkSigner.version())

    UberApkSigner.sign(apks, allow_resign=True, overwrite=True)

    if is_bundle:
        LOG.info('Repacking bundle')
        cache_bundle = os.path.join(USER_DIRECTORIES.user_cache_dir, f'cached.{ext}')
        Bundle.repack(f'{workdir}_bundle', cache_bundle)
        if os.path.isfile(output):
            os.remove(output)
        shutil.copyfile(cache_bundle, output)
    LOG.info('DONE! {}', output)

    ### Clear cache
  
@main.command(help='Inject a shared library (*.so) in a target apk.')
@click.argument('apk')
@click.option('--library', required=True, multiple=True, help='Shared library (*.so) to inject. If multiple are given, only the first one will be automatically loaded.')
@click.option('--include', multiple=True, required=False, help='Extra files to include in the lib folder. Can be used multiple times to include more files.')
@click.option('--activity', required=False, help='Name of the activity to inject into. If unspecified, the main activity will be injected.')
@click.option('--output', required=False, default=os.getcwd(), help='Custom path where the patched apk should be saved.')
@click.option('--override', is_flag=True, help='Override apk in place.')
@click.option('--aapt2', required=False, help='Either to use aapt2 when building the apk or not.')
def inject(apk, library, include, activity, output, override, aapt2):
    Java.install(progress_callback=_install_callback)
    _inject(apk, library, include, activity, output, override, aapt2)


@main.command(help='Inject a frida gadget in a target apk.')
@click.argument('apk')
@click.option('--script', required=False, help='Inject a javascript to be loaded when the gadget starts.')
@click.option('--codeshare', required=False, help='Same as --script but uses Frida Codeshare as the source javascript.')
@click.option('--arch', required=False, help='Architecture to inject for. If empty --all, --adb or --gadget must be specified.')
@click.option('--adb', is_flag=True, help='Use adb (if installed) to detect device architecture.')
@click.option('--activity', required=False, help='Name of the activity to inject into. If unspecified, the main activity will be injected.')
@click.option('--output', required=False, default=os.getcwd(), help='Custom path where the patched apk should be saved.')
@click.option('--override', is_flag=True, help='Override apk in place.')
@click.option('--aapt2', required=False, help='Either to use aapt2 when building the apk or not.')
def frida(apk, script, codeshare, arch, adb, activity, output, override, aapt2):
    if arch:
        if arch not in [ARCH.ARM, ARCH.ARM64, ARCH.X86, ARCH.X64]:
            LOG.error('{} not a supported arch: {}, {}, {}, {}',
                      arch, ARCH.ARM, ARCH.ARM64, ARCH.X86, ARCH.X64)
        Java.install(progress_callback=_install_callback)
        gadget = Gadget.get_gadget_path(arch)
        if not gadget:
            Gadget.update(_install_callback)
            gadget = Gadget.get_gadget_path(arch)
    elif adb:
        Adb.install(None, _install_callback)
        arch = Adb.get_architecture()
        if not arch:
            LOG.info('Waiting for adb device...')
            Adb.wait_for_device()
            arch = Adb.get_architecture()
        if not arch:
            LOG.error(
                'Failed to connect to any device. Make sure you have adb installed, and the device is properly connected.')
        gadget = Gadget.get_gadget_path(arch)
        if not gadget:
            Gadget.update(_install_callback)
            gadget = Gadget.get_gadget_path(arch)
    else:
        LOG.error('--arch or --adb is missing.')
    if os.path.exists(gadget):
        gadget_tmp = os.path.join(
            USER_DIRECTORIES.user_cache_dir, 'libfrida-gadget.so')
        shutil.copy(gadget, gadget_tmp)
    if script:
        if not os.path.isfile(script):
            LOG.error('Not a valid script file: {}', script)
        script_tmp = os.path.join(
            USER_DIRECTORIES.user_cache_dir, 'libhook.js.so')
        shutil.copyfile(script, script_tmp)
        script = [script_tmp,]
    if codeshare:
        if codeshare.endswith('/'):
            uri = codeshare[:-1]
            project_url = f"https://codeshare.frida.re/api/project/{uri}/"
            script = [ os.path.join(
                USER_DIRECTORIES.user_cache_dir, 'libhook.js.so') ]
            download_file(project_url, script[-1], _install_callback)
    if script or codeshare:
        config = {
            "interaction":
            {
                "type": "script",
                "address": "127.0.0.1",
                "port": 27042,
                "path": "./libhook.js.so",
                "on_port_conflict": "pick-next",
                "on_load": "resume"
            }
        }
        config_path = os.path.join(
            USER_DIRECTORIES.user_cache_dir, 'libfrida-gadget.config.so')
        with open(config_path, 'w') as f:
            f.write(json.dumps(config))
        script.append(config_path)
    _inject(apk, [gadget_tmp,], script, activity, output, override, aapt2)


@main.command(help='Apk utils')
@click.argument('apk')
@click.option('--activities', is_flag=True, help='Gets all activities')
@click.option('--permissions', is_flag=True, help='Gets all permissions')
@click.option('--libraries', is_flag=True, help='Gets all libraries')
@click.option('--recievers', is_flag=True, help='Gets all receivers')
def apk(apk, activities, permissions, libraries, recievers):
    def _parse_apk():
        Java.install(progress_callback=_install_callback)
        Apktool.install(progress_callback=_install_callback)

        decoded_dir = os.path.join(USER_DIRECTORIES.user_cache_dir, 'decoded_tmp')

        Apktool.decode(apk, output=decoded_dir, framework_path=f'{decoded_dir}_framework', no_res=True, no_src=True, force_manifest=True)
        manifest_path = os.path.join(decoded_dir, 'AndroidManifest.xml')
        if activities:
            for activity in Manifest.get_activities(manifest_path):
                LOG.info(activity)
        if permissions:
            for permission in Manifest.get_permissions(manifest_path):
                LOG.info(permission)
        if libraries:
            for library in Manifest.get_libraries(manifest_path):
                LOG.info(library)
        if recievers:
            for reciever in Manifest.get_receivers(manifest_path):
                LOG.info('{}', reciever)
    if not os.path.isfile(apk):
        LOG.error('{} - not such file.', apk)
    if Bundle.is_bundle(apk):
        with tempfile.TemporaryDirectory() as tmp:
            apk, base = Bundle.extract(apk, tmp)
            apk = os.path.join(apk, base)
            return _parse_apk()
    _parse_apk()


@main.command(help='Manage plugins')
@click.option('--list', is_flag=True, help='List all installed and available plugins.')
@click.option('--install', required=False, help='Installs a plugin by name.')
@click.option('--remove', required=False, help='Removes an installed plugin.')
def plugins(list, install, remove):
    plugin_path = os.path.join(USER_DIRECTORIES.user_data_dir, "plugins")
    online_plugins = None

    if list:
        response = requests.get(__PLUGIN_DATA__)
        if response.status_code == 200:
            online_plugins = response.json()
        LOG.warning('INSTALLED PLUGINS')
        for plugin in os.listdir(plugin_path):
            if plugin.endswith('.py'):
                LOG.info(plugin[:-3])
        LOG.warning('AVAILABLE PLUGINS')
        if online_plugins:
            for plugin in online_plugins:
                LOG.info(plugin['name'])
                LOG.info('\tauthor: {}', plugin['author'])
                LOG.info('\tdescription: {}', plugin['description'])
                LOG.info('\tversion: {}', plugin['version'])
                LOG.info('\turl: {}', plugin['url'])

    if install:
        response = requests.get(__PLUGIN_DATA__)
        if response.status_code == 200:
            online_plugins = response.json()
            if online_plugins:
                for plugin in online_plugins:
                    if plugin['name'] == install:
                        LOG.info('Installing {} by {} version {}',
                                 plugin['name'], plugin['author'], plugin['version'])
                        response = requests.get(plugin['url'])
                        if response.status_code != 200:
                            LOG.error(
                                'Failed to download {}. Request failed with: {}', install, response.status_code)
                        out = os.path.join(plugin_path, f'{plugin["name"]}.py')
                        with open(out, 'w') as out_f:
                            out_f.write(response.text)
    if remove:
        for plugin in os.listdir(plugin_path):
            if plugin.endswith('.py'):
                name = plugin[:-3]
                if name == remove:
                    path = os.path.join(plugin_path, plugin)
                    os.remove(path)
                    LOG.info('Removed {}', path)


def start():
    plugin_folder = os.path.join(USER_DIRECTORIES.user_data_dir, 'plugins')
    if not os.path.exists(plugin_folder):
        os.makedirs(plugin_folder)
    _PluginLoader(plugin_folder)
    main()


if __name__ == '__main__':
    start()



from xml.dom.minidom import parseString

from typing import List

class Manifest:
    """
    AndroidManifest.xml utils used across the project
    """
    @staticmethod
    def get_activities(manifest : str) -> List[str]:
        """
        Get a list of activities declared in the target manifest.

        :param manifest: Path to the decoded AndroidManifest.xml.
        :type manifest: str
        :return: A sorted list of activities.
        :rtype: List[str]
        """
        data = ""
        with open(manifest, 'r') as m:
            data = m.read()
        dom = parseString(data)
        nodes = dom.getElementsByTagName('activity')
        activities = [node.getAttribute("android:name") for node in nodes]
        return sorted(activities)
    @staticmethod
    def get_main_activity(manifest : str) -> str:
        """
        Get the main activity declared in the target manifest.

        :param manifest: Path to the decoded AndroidManifest.xml.
        :type manifest: str
        :return: The name of the main activity
        :rtype: str
        """
        main_activity = None
        data = ""
        with open(manifest, 'r') as m:
            data = m.read()
        dom = parseString(data)
        nodes = dom.getElementsByTagName('activity')
        for node in nodes:
            intent_filters = node.getElementsByTagName('intent-filter')
            if intent_filters:
                for intent_filder in intent_filters:
                    actions = intent_filder.getElementsByTagName('action')
                    categories = intent_filder.getElementsByTagName('category')
                    for action in actions:
                        if action.getAttribute('android:name') == 'android.intent.action.MAIN':
                            for category in categories:
                                if category.getAttribute('android:name') == 'android.intent.category.LAUNCHER':
                                    main_activity = node.getAttribute("android:name")
        return main_activity

    @staticmethod
    def get_permissions(manifest : str) -> List[str]:
        """
        Get a list of permissions declared in the target manifest.

        :param manifest: Path to the decoded AndroidManifest.xml.
        :type manifest: str
        :return: A sorted list of permissions.
        :rtype: List[str]
        """
        data = ""
        with open(manifest, 'r') as m:
            data = m.read()
        dom = parseString(data)
        nodes = dom.getElementsByTagName('uses-permission')
        nodes.extend(dom.getElementsByTagName('uses-permission-sdk-23'))
        permissions = [node.getAttribute("android:name") for node in nodes]
        return sorted(permissions)
    @staticmethod
    def get_libraries(manifest : str) -> List[str]:
       """
        Get a list of activities declared in the target manifest.

        :param manifest: Path to the decoded AndroidManifest.xml.
        :type manifest: str
        :return: A sorted list of activities.
        :rtype: List[str]
        """
       data = ""
       with open(manifest, 'r') as m:
        data = m.read()
        dom = parseString(data)
        nodes = dom.getElementsByTagName('uses-library')
        libraries = [node.getAttribute("android:name") for node in nodes]
        return sorted(libraries)


import os
from typing import Callable, List, Optional

from . import __UBER_SIGNER_VERSION__, DEPENDENCIES, USER_DIRECTORIES
from .download import download_file
from .java import Java


def _download_uberapksigner(progress_callback=None):
    path = os.path.join(USER_DIRECTORIES.user_data_dir,
                        f'uber-apk-signer-{__UBER_SIGNER_VERSION__}.jar')
    if not os.path.isfile(path):
        uri = f'https://github.com/patrickfav/uber-apk-signer/releases/download/v{__UBER_SIGNER_VERSION__}/uber-apk-signer-{__UBER_SIGNER_VERSION__}.jar'
        return download_file(uri, path, progress_callback)
    return path


class UberApkSigner:
    """
    Execute uber-apk-signer commands
    """
    @staticmethod
    def version():
        """
        Obtain the version of uber-apk-signer.

        :return: The stdout from the command execution.
        :rtype: str
        """
        return Java.run_jar(DEPENDENCIES.uber_apk_signer, '--version')

    @staticmethod
    def sign(apks: List[str], allow_resign: Optional[bool] = False, debug: Optional[bool] = False,
             dry_run: Optional[bool] = False, ks: Optional[str] = None, ks_alias: Optional[str] = None,
             ks_debug: Optional[str] = None, ks_key_pass: Optional[str] = None, ks_pass: Optional[str] = None,
             lineage: Optional[str] = None, out: Optional[str] = None, overwrite: Optional[bool] = False,
             skip_zip_align: Optional[bool] = False, only_verify: Optional[bool] = False,
             zip_align_path: Optional[str] = None, verify_sha256: Optional[str] = None) -> str:
        """
        Sign an APK or a set of APKs.

        :return: The stdout from the command execution.
        :rtype: str
        """

        command = '-a {}'.format(' '.join(apks))

        if allow_resign:
            command += " --allowResign"

        if debug:
            command += " --debug"

        if dry_run:
            command += " --dryRun"

        if ks:
            command += f' --ks {ks}'

        if ks_alias:
            command += f' --ksAlias {ks_alias}'

        if ks_debug:
            command += f' --ksDebug {ks_debug}'

        if ks_key_pass:
            command += f' --ksKeyPass {ks_key_pass}'

        if ks_pass:
            command += f' --ksPass {ks_pass}'

        if lineage:
            command += f' -l {lineage}'

        if out:
            command += f' -o {out}'

        if overwrite:
            command += " --overwrite"

        if skip_zip_align:
            command += " --skipZipAlign"

        if only_verify:
            command += " -y"

        if zip_align_path:
            command += f' --zipAlignPath {zip_align_path}'

        if verify_sha256:
            command += f' --verifySha256 {verify_sha256}'

        return Java.run_jar(DEPENDENCIES.uber_apk_signer, command)

    @staticmethod
    def install(path: str = None, progress_callback: Callable = None) -> None:
        """
        Install uber-apk-signer.

        :param path: Path to existing uber-apk-signer.jar. If not found, will use the system installed one or download it. Defaults to None.
        :type path: str, optional
        :param progress_callback: Callback to be called when install progress changes, defaults to None.
        :type progress_callback: callable, optional
        """
        DEPENDENCIES.add_dependency(
            'uber-apk-signer', path=path, fallback=_download_uberapksigner, fallback_args=(progress_callback,), use_path=False)



import logging
import os

from appdirs import AppDirs

from .dependencies import DependenciesManager
from .logger import ApkInjectorLogger

LOG = ApkInjectorLogger(__name__)

USER_DIRECTORIES: AppDirs = AppDirs(__name__, 'nitanmarcel')

if not os.path.isdir(USER_DIRECTORIES.user_data_dir):
    os.makedirs(USER_DIRECTORIES.user_data_dir)
if not os.path.isdir(USER_DIRECTORIES.user_cache_dir):
    os.makedirs(USER_DIRECTORIES.user_cache_dir)


DEPENDENCIES = DependenciesManager(USER_DIRECTORIES.user_data_dir)

__APKTOOL_VERSION__ = '2.9.0'
__UBER_SIGNER_VERSION__ = '1.3.0'
__JAVA_VERSION__ = '17'

__PLUGIN_DATA__ = 'https://raw.githubusercontent.com/nitanmarcel/apkinjector/main/plugins/plugins.json'

if DEPENDENCIES.has_missing_required_dependency():
    LOG.error('Some dependencies are missing.')



from typing import Union

from .arch import ARCH


def abi_to_arch(abi: str) -> Union[ARCH, None]:
    """Convert abi string to architecture

    Args:
        abi (str): Abi string

    Returns:
        Union[ARCH, None]: Converted abi to apkinjector.arch.Arch or None if the abi is not recognized.
    """
    mappings = {
        'armeabi-v7a': ARCH.ARM,
        'arm64_v8a': ARCH.ARM64,
        'x86': ARCH.X86,
        'x86_64': ARCH.X64,
        'arm64-v8a': ARCH.ARM64,
        'armeabi_v7a': ARCH.ARM
    }
    if abi in mappings.keys():
        return mappings[abi]
    return None


def arch_to_abi(arch: ARCH) -> Union[str, None]:
    """Convert apkinjector.arch.ARCH string to abi

    Args:
        arch (str): apkinjector.arch.ARCH

    Returns:
        Union[ARCH, None]: Converted apkinjector.arch.ARCH to abi string or None if the apkinjector.arch.ARCH is not recognized.
    """
    mappings = {
        ARCH.ARM: 'armeabi-v7a',
        ARCH.ARM64: 'arm64-v8a',
        ARCH.X86: 'x86',
        ARCH.X64: 'x86_64'
    }
    if arch in mappings.keys():
        return mappings[arch]
    return None



import subprocess

import requests

from . import DEPENDENCIES


class Frida:
    """
    A utility class for running Frida commands.
    """

    @staticmethod
    def version():
        """
        Obtain the version of frida.

        :return: The stdout from the command execution.
        :rtype: str
        """
        if DEPENDENCIES.frida is not None:
            process = subprocess.run(
                f'{DEPENDENCIES.frida} --version', shell=True, capture_output=True, text=True)
            return process.stdout.strip()
        uri = f"https://api.github.com/repos/frida/frida/releases/latest"
        response = requests.get(uri)
        js = response.json()
        return js['tag_name']


DEPENDENCIES.add_dependency('frida', required=False)



from dataclasses import dataclass
from typing import Callable, Optional

import os
import shutil
import requests

from . import LOG, USER_DIRECTORIES


@dataclass
class ProgressDownloading:
    filename: str
    progress: int


@dataclass
class ProgressFailed:
    filename: str
    status_code: int


@dataclass
class ProgressCompleted:
    filename: str
    path: str

@dataclass
class ProgressUnknown:
    filename: str
    path: str


def download_file(url: str, target_path: str, progress_callback: Optional[Callable] = None) -> str:
    """
    Download a file from a given URL and save it on the target path.

    :param url: The URL from where file needs to be downloaded.
    :type url: str
    :param target_path: The path on local system where the downloaded file should be saved.
    :type target_path: str
    :param progress_callback: Function to call when the progress changes, defaults to None.
    :type progress_callback: Optional[Callable]
    :return: The target path where the file has been downloaded.
    :rtype: str
    """
    def _callable(args):
        if progress_callback is not None:
            progress_callback(args)
    file_name = target_path.split('/')[-1]
    tmp_path = os.path.join(USER_DIRECTORIES.user_cache_dir, "downloads")
    if not os.path.exists(tmp_path):
        os.mkdir(tmp_path)

    tmp_path = os.path.join(tmp_path, file_name)

    if os.path.isfile(tmp_path):
        os.remove(tmp_path)
    response = requests.get(url, stream=bool(progress_callback))
    total_length = response.headers.get('content-length')
    total_length = int(total_length)

    if response.status_code != 200:
        _callable(ProgressFailed(filename=file_name,
                  status_code=response.status_code))
        return
    _callable(ProgressDownloading(filename=file_name, progress=0))
    with open(tmp_path, 'wb') as f:
        if not bool(progress_callback):
            f.write(response.content)
        else:
            downloaded = 0
            for chunk in response.iter_content(chunk_size=1020):
                if chunk:
                    downloaded += len(chunk)
                    percentage = int(downloaded * 100 / total_length)
                    _callable(ProgressDownloading(file_name, percentage))
                    f.write(chunk)
    if os.path.isfile(tmp_path):
        shutil.move(tmp_path, target_path)
    _callable(ProgressCompleted(file_name, target_path))

    return target_path




from typing import List, Union
import json
import re
import os
import zipfile

from dataclasses import dataclass
from .arch import ARCH
from .utils import arch_to_abi

@dataclass
class ConfigArch:
    """
    A Data class representing the architecture information in a split APK file.
    
    :param str file_name: The name of the APK file.
    :param str arch: The architecture type of the APK file ('ARM', 'ARM64', 'X86', 'X64').
    """

    file_name: str
    arch: str

@dataclass
class ConfigLocale:
    """
    A Data class representing the locale information in a split APK file.

    :param str file_name: A string representing the name of the APK file.
    :param str locale: A string indicating the locale config of the APK file.
    """

    file_name: str
    locale: str

@dataclass
class ConfigDpi:
    """
    A Data class representing the DPI information in a split APK file.

    :param str file_name: The name of the APK file.
    :param str dpi: A string representing the screen DPI config of the APK.
    """

    file_name: str
    dpi: str

@dataclass
class BundleInfo:
    """
    A data class representing general information about an Android Bundle file.

    :param str base_apk: The name of the base APK file extracted from the Android Bundle.
    :param List[Union[ConfigDpi, ConfigLocale, ConfigArch]] configs: Configurations of the APK files extracted from the Android Bundle.
    """

    base_apk: str
    configs: List[Union[ConfigDpi, ConfigLocale, ConfigArch]]

class Bundle:
    """
    Tools to work with apk bundles (xapk and apks)
    """
    @staticmethod
    def extract(source: str, output_path: str) -> BundleInfo:
        """
        Extract a bundle (xapk and apks) to a given path.

        :param source: Path to the bundle.
        :type source: str
        :param output_path: Destination of the bundled.
        :type output_path: str
        :return: BundleInfo class containing the name of the base apk and configuration ifno.
        :rtype: BundleInfo
        """
        name = os.path.basename(source)[:-5]
        base_apk = None
        if os.path.exists(output_path):
            output_path = os.path.join(output_path, name)
        os.makedirs(output_path)
        with zipfile.ZipFile(source) as zip:
            zip.extractall(output_path)
        sai_manifestv1 = os.path.join(output_path, 'meta.sai_v1.json')
        sai_manifestv2 = os.path.join(output_path, 'meta.sai_v2.json')
        apkm_installer_url = os.path.join(output_path, 'APKM_installer.url')
        base_apk = None
        configs = []
        matches = [None]

        if os.path.isfile(sai_manifestv1) or os.path.isfile(sai_manifestv2) or os.path.isfile(apkm_installer_url):       
            base_apk = 'base.apk'
        else:
            manifest = os.path.join(output_path, 'manifest.json')
            with open(manifest, 'r') as m_file:
                manifest = json.load(m_file)
                for apk in manifest['split_apks']:
                    if apk['id'] == 'base':
                        base_apk = apk['file']
        files = os.listdir(output_path)
        for file in files:
            if file.endswith('.apk'):
                if f'.{arch_to_abi(ARCH.ARM64)}.' in file and file not in matches:
                    configs.append(ConfigArch(file, ARCH.ARM64))
                    matches.append(file)
                if f'.{arch_to_abi(ARCH.ARM)}.' in file and file not in matches:
                    configs.append(ConfigArch(file, ARCH.ARM))
                    matches.append(file)

                if f'.{arch_to_abi(ARCH.ARM64).replace("-", "_")}.' in file and file not in matches:
                    configs.append(ConfigArch(file, ARCH.ARM64))
                    matches.append(file)
                if f'.{arch_to_abi(ARCH.ARM).replace("-", "_")}.' in file and file not in matches:
                    configs.append(ConfigArch(file, ARCH.ARM))
                    matches.append(file)

                if f'.{arch_to_abi(ARCH.X64)}.' in file and file not in matches:
                    configs.append(ConfigArch(file, ARCH.X64))
                    matches.append(file)
                if f'.{arch_to_abi(ARCH.X86)}.' in file and file not in matches:
                    configs.append(ConfigArch(file, ARCH.X86))
                    matches.append(file)
                        
                # Check for locale
                locale_match = re.search(r'config\.([a-z]{2}(?:_r[A-Z]{2})?)\.apk', file)
                if locale_match and file not in matches:
                    configs.append(ConfigLocale(file, locale_match.group(1)))
                
                # Check for dpi
                dpi_match = re.search(r'config\.(.*?)\.apk', file)
                if dpi_match and dpi_match.group(1).endswith('dpi') and file not in matches:
                    configs.append(ConfigDpi(file, dpi_match.group(1).replace('dpi', '')))
                    matches.append(file)
        info = BundleInfo(base_apk, configs)
        return info

    @staticmethod
    def repack(source: str, output_path: str) -> str:
        """
        Repacks a bundle (xapk and apks) to a given path.

        :param source: Path to the extracted bundle.
        :type source: str
        :param output_path: Path to where to save the bundle file.
        :type output_path: str
        :return: Location of the repacked bundle.
        """
        with zipfile.ZipFile(output_path, 'w', zipfile.ZIP_DEFLATED) as zip:
            for foldername, subfolders, filenames in os.walk(source):
                for filename in filenames:
                    abspath = os.path.join(foldername, filename)
                    relpath = os.path.relpath(abspath, source)
                    zip.write(abspath, arcname=relpath)
        return output_path

    @staticmethod
    def is_bundle(source: str) -> bool:
        """
        Checks if the source is a bundle or not.

        :param source: Source to check.
        :type source: str
        :return: True if it's a bundle, False otherwise.
        :rtype: bool
        """
        name = os.path.basename(source)
        return name.endswith('.xapk') or name.endswith('.apks') or name.endswith('.apkm')
